# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dataset_serialize.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='dataset_serialize.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x17\x64\x61taset_serialize.proto\"\x81\x01\n\x08\x44setFile\x12\r\n\x05parms\x18\x01 \x01(\t\x12\x10\n\x08\x65xamples\x18\x02 \x03(\t\x12%\n\x06images\x18\x03 \x03(\x0b\x32\x15.DsetFile.ImagesEntry\x1a-\n\x0bImagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x62\x06proto3'
)




_DSETFILE_IMAGESENTRY = _descriptor.Descriptor(
  name='ImagesEntry',
  full_name='DsetFile.ImagesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='DsetFile.ImagesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='DsetFile.ImagesEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=112,
  serialized_end=157,
)

_DSETFILE = _descriptor.Descriptor(
  name='DsetFile',
  full_name='DsetFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='parms', full_name='DsetFile.parms', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='examples', full_name='DsetFile.examples', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='images', full_name='DsetFile.images', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DSETFILE_IMAGESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=157,
)

_DSETFILE_IMAGESENTRY.containing_type = _DSETFILE
_DSETFILE.fields_by_name['images'].message_type = _DSETFILE_IMAGESENTRY
DESCRIPTOR.message_types_by_name['DsetFile'] = _DSETFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DsetFile = _reflection.GeneratedProtocolMessageType('DsetFile', (_message.Message,), {

  'ImagesEntry' : _reflection.GeneratedProtocolMessageType('ImagesEntry', (_message.Message,), {
    'DESCRIPTOR' : _DSETFILE_IMAGESENTRY,
    '__module__' : 'dataset_serialize_pb2'
    # @@protoc_insertion_point(class_scope:DsetFile.ImagesEntry)
    })
  ,
  'DESCRIPTOR' : _DSETFILE,
  '__module__' : 'dataset_serialize_pb2'
  # @@protoc_insertion_point(class_scope:DsetFile)
  })
_sym_db.RegisterMessage(DsetFile)
_sym_db.RegisterMessage(DsetFile.ImagesEntry)


_DSETFILE_IMAGESENTRY._options = None
# @@protoc_insertion_point(module_scope)



"""
SiliconLife Eyeflow
Log singleton object

Author: Alex Sobral de Freitas
"""

import os
import sys
import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
import yaml
#----------------------------------------------------------------------------------------------------------------------------------

if "CONF_PATH" not in os.environ:
    os.environ["CONF_PATH"] = "/opt/eyeflow/conf"

with open(os.path.join(os.environ["CONF_PATH"], "eyeflow_conf.yaml"), "r") as ymlfile:
    CONFIG = yaml.safe_load(ymlfile)
#----------------------------------------------------------------------------------------------------------------------------------

class LogObj:
    """
    Singletron class to operate log
    """

    class __LogObj:
        def __init__(self):
            """
            Initialize log
            """
            self.application_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]

            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(logging.INFO)

            log_path = Path(CONFIG["log"]["log_folder"])

            if not log_path.is_dir():
                log_path.mkdir(parents=True, exist_ok=True)

            log_file = Path(log_path, f"{self.application_name}.log")
            print("Log File: %s" % str(log_file))

            try:
                log_file_handler = RotatingFileHandler(
                    filename=str(log_file),
                    mode='a',
                    maxBytes=(10 * 1024 * 1024),
                    backupCount=5
                    )
            except:
                raise IOError("Couldn't create/open file \"" + str(log_file) + "\". Check permissions.")

            log_file_handler.setFormatter(logging.Formatter('%(asctime)s - %(funcName)s - %(levelname)s - %(message)s'))
            log_file_handler.setLevel(logging.INFO)
            self.logger.addHandler(log_file_handler)

            # create console handler and set level to debug
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
            console_handler.setLevel(logging.DEBUG)
            self.logger.addHandler(console_handler)

        def __getattr__(self, name):
            return getattr(self.logger, name)

    instance = None

    def __init__(self):
        if not LogObj.instance:
            LogObj.instance = LogObj.__LogObj()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __str__(self):
        return 'Log singletron object'
#----------------------------------------------------------------------------------------------------------------------------------

log = LogObj()
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Class for log batch of extracted images from detection

Author: Alex Sobral de Freitas
"""

import os

import json
import datetime
import random
from bson import ObjectId
import cv2

from eyeflow_sdk.file_access import FileAccess
from eyeflow_sdk.log_obj import log
#----------------------------------------------------------------------------------------------------------------------------------

MAX_EXTRACT_FILES = 800

def clear_log(extract_path, max_files=MAX_EXTRACT_FILES):
    files_list = os.listdir(extract_path)
    if len(files_list) > max_files:
        date_list = [(filename, datetime.datetime.fromtimestamp(os.path.getmtime(os.path.join(extract_path, filename)))) for filename in files_list]
        exclude_list = sorted(date_list, key=lambda x: x[1])[:len(files_list) - max_files]
        for filename, _ in exclude_list:
            try:
                os.remove(os.path.join(extract_path, filename))
            except:
                pass
#----------------------------------------------------------------------------------------------------------------------------------

def upload_extracts(dataset_name, cloud_parms):
    """
    Upload extracts of process to cloud
    """

    log.info(f"Upload extracts dataset: {dataset_name}")

    def save_extract_list(extract_path):

        date_list = []
        for filename in os.listdir(extract_path):
            if filename.endswith('_data.json'):
                try:
                    filepath = os.path.join(extract_path, filename)
                    with open(filepath, 'r') as json_file:
                        data = json.load(json_file)
                        if 'frame_time' in data:
                            frame_time = data['frame_time']
                        else:
                            frame_time = datetime.datetime.fromtimestamp(os.path.getmtime(filepath))
                        date_list.append([filename, frame_time])
                except:
                    pass

        files_list = sorted(date_list, key=lambda x: x[1], reverse=True)
        with open(os.path.join(extract_path, 'extract_files.json'), 'w', newline='', encoding='utf8') as file_p:
            json.dump({"extract_list": files_list}, file_p, ensure_ascii=False, indent=2, default=str)

    file_ac = FileAccess(storage="extract", resource_id=dataset_name, cloud_parms=cloud_parms)
    clear_log(file_ac.get_local_folder())
    save_extract_list(file_ac.get_local_folder())
    file_ac.sync_files(origin="local")
#----------------------------------------------------------------------------------------------------------------------------------

class VideoLog(object):
    def __init__(self, dataset_name, vlog_size):
        self._vlog_size = vlog_size
        file_ac = FileAccess(storage="extract", resource_id=dataset_name)
        self._dest_path = file_ac.get_local_folder()
        self._last_log = datetime.datetime(2000, 1, 1)


    def log_batch(self, image_batch, output_batch, annotations):
        for idx, image in enumerate(image_batch):
            if random.random() < float(self._vlog_size):
                obj_id = str(ObjectId())
                cv2.imwrite(os.path.join(self._dest_path, obj_id + '.jpg'), image[0])
                img_data = {
                    "_id": obj_id,
                    "date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                    "img_height": image[0].shape[0],
                    "img_width": image[0].shape[1],
                    "detections": annotations[idx],
                    "annotations": annotations[idx]
                }

                if 'frame_time' in image[2]:
                    img_data['frame_time'] = image[2]['frame_time']

                with open(os.path.join(self._dest_path, obj_id + '_data.json'), 'w', newline='', encoding='utf8') as file_p:
                    json.dump(img_data, file_p, ensure_ascii=False, indent=2, default=str)

                if (datetime.datetime.now() - self._last_log) > datetime.timedelta(minutes=1):
                    clear_log(self._dest_path)
                    self._last_log = datetime.datetime.now()
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Video manipulation functions

Author: Alex Sobral de Freitas
"""

import os
import json
import datetime
from bson.objectid import ObjectId
import cv2
#---------------------------------------------------------------------------

class VideoPlay():
    def __init__(self, filename, window_size): #, video_scale):
        self._window_size = window_size
        #self._frame_scale = video_scale
        self._video_cap = cv2.VideoCapture(filename)

        self._video_frames = {}
        self._min_frame = 0
        self._max_frame = 0
        self._video_open = True

        success, video_frame = self._video_cap.read()
        if not success:
            raise Exception("Fail to read video: {}".format(filename))

        self._video_frames[str(self._min_frame)] = video_frame

    def is_open(self):
        return self._video_open


    def get_total_frames(self):
        return int(self._video_cap.get(cv2.CAP_PROP_FRAME_COUNT))


    def get_frame_rate(self):
        return int(self._video_cap.get(cv2.CAP_PROP_FPS))


    def get_frame(self, frame):
        if frame < self._min_frame:
            return self._video_frames[str(self._min_frame)], self._min_frame
        elif frame >= self._max_frame:
            while self._max_frame < frame:
                if self._video_open:
                    success, video_frame = self._video_cap.read()
                    if not success:
                        self._video_open = False
                        return self._video_frames[str(self._max_frame)], self._max_frame
                    else:
                        self._max_frame += 1
                        self._video_frames[str(self._max_frame)] = video_frame

        self.discard_frames_window()
        return self._video_frames[str(frame)], frame


    def discard_frames_window(self):
        if (self._max_frame - self._min_frame) < self._window_size:
            return

        while (self._max_frame - self._min_frame) > self._window_size:
            del self._video_frames[str(self._min_frame)]
            self._min_frame += 1


    def extract_frames(self, num_frames, dest_path):
        f_max = self.get_total_frames()
        f_count = 0
        f_step = f_max // num_frames
        frame = f_step
        while f_count <= num_frames:
            img, frame = self.get_frame(frame)
            img_name = str(ObjectId())
            cv2.imwrite(os.path.join(dest_path, img_name + '.jpg'), img)
            img_data = {
                "date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                "img_height": img.shape[0],
                "img_width": img.shape[1],
                "detections": {"instances":[]},
                "annotations": {"instances":[]}
            }

            with open(os.path.join(dest_path, img_name + '_data.json'), 'w', newline='', encoding='utf8') as fp:
                json.dump(img_data, fp, ensure_ascii=False, indent=2)

            frame += f_step
            f_count += 1
#----------------------------------------------------------------------------------------------------------------------------------

def get_video_info(video_path, video_file):
    # log.info("Get video info {}".format(video_file))

    file = os.path.join(video_path, video_file)
    if not os.path.isfile(file):
        raise Exception('Video not found: ' + file)

    vid_cap = VideoPlay(file, 300)

    total_frames = vid_cap.get_total_frames()
    total_time = round(total_frames / vid_cap.get_frame_rate(), 2)

    video_frame, _ = vid_cap.get_frame(0)

    frame_height = video_frame.shape[0]
    frame_width = video_frame.shape[1]

    video_info = {
        "video_file": video_file,
        "total_frames": total_frames,
        "total_time": total_time,
        "frame_height": frame_height,
        "frame_width": frame_width
    }

    return json.dumps(video_info)
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Class for access files synchronized with cloud storage

Author: Alex Sobral de Freitas
"""

import os
from pathlib import Path
import importlib
import datetime
import time
import traceback

from eyeflow_sdk.log_obj import log, CONFIG
#----------------------------------------------------------------------------------------------------------------------------------

class FileAccess():
    """
    Class for handle files cloud synchronized transparently
    """
    def __init__(self, storage, resource_id=None, cloud_parms=None):
        config_parms = CONFIG["file-service"]

        if storage not in config_parms:
            raise Exception(f"Unknown storage: {storage}")

        self.storage = storage
        self.resource_id = resource_id

        if resource_id is not None:
            self._local_folder = config_parms[storage]["local_folder"] + "/" + resource_id
        else:
            self._local_folder = config_parms[storage]["local_folder"]

        self._cloud_folder = None
        if "cloud_folder" in config_parms[storage]:
            self._cloud_folder = config_parms[storage]["cloud_folder"]

        # make folder if not exists
        folder_path = Path(self._local_folder)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        self._cloud_obj = None
        if cloud_parms is not None:
            comp_lib = importlib.import_module(f'cloud_store.{cloud_parms["provider"]}')
            self._cloud_obj = comp_lib.Connector(**cloud_parms)

    @staticmethod
    def _get_list_files_info(folder, prefix='', sufix=''):
        file_list = []
        for filename in os.listdir(folder):
            if os.path.isfile(os.path.join(folder, filename)) \
               and filename.startswith(prefix) and filename.endswith(sufix):
                file_stat = os.stat(os.path.join(folder, filename))
                file_info = {"filename": filename,
                             "creation_date": datetime.datetime.fromtimestamp(file_stat.st_mtime, datetime.timezone.utc),
                             "modified_date": datetime.datetime.fromtimestamp(os.path.getmtime(os.path.join(folder, filename)), datetime.timezone.utc),
                             "file_size": file_stat.st_size
                            }

                file_list.append(file_info)
            elif os.path.isdir(os.path.join(folder, filename)):
                subfolder = os.path.join(folder, filename)
                subfolder_files = FileAccess._get_list_files_info(subfolder, prefix, sufix)
                for finfo in subfolder_files:
                    finfo["filename"] = os.path.join(os.path.split(subfolder)[1], finfo["filename"])

                file_list.extend(subfolder_files)

        return file_list


    def sync_files(self, origin, file_list=None):
        """
        Synchronizes files between local and cloud folder
        If file_list then synchronize only files in this list
        """
        if self._cloud_obj is None:
            raise Exception("Cloud provider not defined")

        if self._cloud_folder is None:
            raise Exception("Cloud folder not defined")

        if origin not in ["cloud", "local", "both"]:
            raise Exception(f"'origin'={origin} must be one of: ('cloud', 'local', 'both')")

        try:
            local_files = self._get_list_files_info(self._local_folder)
            cloud_files = self._cloud_obj.list_files_info(folder=self._cloud_folder, resource_id=self.resource_id)

            if origin in ["cloud", "both"]:
                download_files = []
                for c_file in cloud_files:
                    if file_list is not None and c_file["filename"] not in file_list:
                        continue

                    for l_file in local_files:
                        if l_file["filename"] == c_file["filename"] and \
                        l_file["file_size"] == c_file["file_size"] and \
                        l_file["modified_date"] >= c_file["modified_date"]:
                            break
                    else:
                        download_files.append(c_file)

                if download_files:
                    log.info(f"Storage: {self.storage} - Resource-ID: {self.resource_id} - Download {len(download_files)} files")
                    for c_file in download_files:
                        file_data = self._cloud_obj.download_file(
                            folder=self._cloud_folder,
                            resource_id=self.resource_id,
                            filename=c_file["filename"]
                        )

                        file_path = os.path.join(self._local_folder, c_file["filename"])
                        dirname = os.path.dirname(file_path)
                        if not os.path.isdir(dirname):
                            os.makedirs(dirname, exist_ok=True)

                        with open(file_path, "wb") as fp:
                            fp.write(file_data)

            if origin in ["local", "both"]:
                upload_files = []
                for l_file in local_files:
                    if file_list is not None and l_file["filename"] not in file_list:
                        continue

                    for c_file in cloud_files:
                        if l_file["filename"] == c_file["filename"] and \
                        l_file["file_size"] == c_file["file_size"] and \
                        l_file["modified_date"] <= c_file["modified_date"]:
                            break
                    else:
                        upload_files.append(l_file)

                if upload_files:
                    log.info(f"Storage: {self.storage} - Resource-ID: {self.resource_id} - Upload {len(upload_files)} files")
                    for l_file in upload_files:
                        file_path = os.path.join(self._local_folder, l_file["filename"])
                        with open(file_path, "rb") as fp:
                            file_data = fp.read()

                        if self._cloud_obj.is_file(
                                folder=self._cloud_folder,
                                resource_id=self.resource_id,
                                filename=l_file["filename"],
                            ):
                            self._cloud_obj.delete_file(
                                folder=self._cloud_folder,
                                resource_id=self.resource_id,
                                filename=l_file["filename"],
                            )

                        file_data = self._cloud_obj.upload_file(
                            folder=self._cloud_folder,
                            resource_id=self.resource_id,
                            filename=l_file["filename"],
                            data=file_data
                        )

                        # Set local file modified time to now() to prevent download the same file after
                        date = datetime.datetime.now()
                        mod_time = time.mktime(date.timetuple())
                        try:
                            os.utime(file_path, (mod_time, mod_time))
                        except:
                            pass

        except Exception as excp:
            log.error(traceback.format_exc())
            log.error(f"sync_files error: {excp}")
            raise excp


    def purge_files(self, max_files=800, max_days=None):
        """
        Purge local and cloud files to max_files from de older to newer
        """
        local_files = self._get_list_files_info(self._local_folder)
        if len(local_files) > max_files:
            date_list = [(l_file["filename"], l_file["modified_date"]) for l_file in local_files]
            exclude_list = sorted(date_list, key=lambda x: x[1])[:len(local_files) - max_files]
            log.info(f"Purge local files: {len(exclude_list)}")
            for filename, _ in exclude_list:
                try:
                    os.remove(os.path.join(self._local_folder, filename))
                except:
                    pass

        if self._cloud_obj is None or self._cloud_folder is None:
            return

        cloud_files = self._cloud_obj.list_files_info(folder=self._cloud_folder, resource_id=self.resource_id)
        if len(cloud_files) > max_files:
            date_list = [(l_file["filename"], l_file["modified_date"]) for l_file in cloud_files]
            exclude_list = sorted(date_list, key=lambda x: x[1])[:len(cloud_files) - max_files]
            log.info(f"Purge cloud files: {len(exclude_list)}")
            for filename, _ in exclude_list:
                try:
                    self._cloud_obj.delete_file(
                        folder=self._cloud_folder,
                        resource_id=self.resource_id,
                        filename=filename,
                    )
                except:
                    pass


    def get_file_path(self, filename):
        return os.path.join(self._local_folder, filename)


    def get_local_folder(self):
        return self._local_folder


    def open(self, filename, *args, **kwargs):
        """
        Returns a file pointer to a local file
        """
        file_path = os.path.join(self._local_folder, filename)
        return open(file_path, *args, **kwargs)


    def is_file(self, filename):
        """
        Returns true if file exists
        """
        file_path = os.path.join(self._local_folder, filename)
        return os.path.isfile(file_path)


    def remove_file(self, filename):
        """
        Remove a local file
        """
        file_path = os.path.join(self._local_folder, filename)
        if os.path.isfile(file_path):
            os.remove(file_path)
#----------------------------------------------------------------------------------------------------------------------------------



"""
    Functions to interact with NVidia GPUs
"""

import arrow
import psutil
import pynvml as nv
import subprocess
import time
import xmltodict

from contextlib import contextmanager

from pynvml import nvmlInit, nvmlShutdown

from eyeflow_sdk.log_obj import log
#----------------------------------------------------------------------------------------------------------------------------------

@contextmanager
def nvml_context():
    nvmlInit()
    yield
    nvmlShutdown()
#----------------------------------------------------------------------------------------------------------------------------------


def parse_cmd_roughly(args):
    cmdline = ' '.join(args)
    if 'python -m ipykernel_launcher' in cmdline:
        return 'jupyter'
    python_script = [arg for arg in args if arg.endswith('.py')]
    if len(python_script) > 0:
        return python_script[0]
    else:
        return cmdline if len(cmdline) <= 25 else cmdline[:25] + '...'
#----------------------------------------------------------------------------------------------------------------------------------


def device_status(device_index):
    handle = nv.nvmlDeviceGetHandleByIndex(device_index)
    device_name = nv.nvmlDeviceGetName(handle)
    device_name = device_name.decode('UTF-8')
    nv_procs = nv.nvmlDeviceGetComputeRunningProcesses(handle)
    utilization = nv.nvmlDeviceGetUtilizationRates(handle).gpu
    clock_mhz = nv.nvmlDeviceGetClockInfo(handle, nv.NVML_CLOCK_SM)
    temperature = nv.nvmlDeviceGetTemperature(handle, nv.NVML_TEMPERATURE_GPU)
    pids = []
    users = []
    dates = []
    cmd = None
    for nv_proc in nv_procs:
        pid = nv_proc.pid
        pids.append(pid)
        try:
            proc = psutil.Process(pid)
            users.append(proc.username())
            dates.append(proc.create_time())
            if cmd is None:
                cmd = parse_cmd_roughly(proc.cmdline())
        except psutil.NoSuchProcess:
            users.append('?')
    return {
        'type': device_name,
        'is_available': len(pids) == 0,
        'pids': ','.join([str(pid) for pid in pids]),
        'users': ','.join(users),
        'running_since': arrow.get(min(dates)).humanize() if len(dates) > 0 else None,
        'utilization': utilization,
        'clock_mhz': clock_mhz,
        'temperature': temperature,
        'cmd': cmd,
        }
#----------------------------------------------------------------------------------------------------------------------------------


def device_statuses():
    with nvml_context():
        device_count = nv.nvmlDeviceGetCount()
        return [device_status(device_index) for device_index in range(device_count)]
#----------------------------------------------------------------------------------------------------------------------------------


def _run_cmd(cmd):
    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result.returncode != 0:
        raise Exception(f"Fail executing command: {cmd} - {result.stderr}")

    return result.stdout.decode("utf-8")
#----------------------------------------------------------------------------------------------------------------------------------


def gpu_info():
    gpu_info = xmltodict.parse(_run_cmd(['nvidia-smi', '--query', '--xml-format']))["nvidia_smi_log"]["gpu"]

    if not isinstance(gpu_info, list):
        gpu_info = [gpu_info]

    for idx,_ in enumerate(gpu_info):
        mem_used = int(gpu_info[idx]["fb_memory_usage"]["used"].strip().replace('MiB', ''))
        mem_total = int(gpu_info[idx]["fb_memory_usage"]["total"].strip().replace('MiB', ''))
        gpu_info[idx]['index'] = str(idx)
        gpu_info[idx]['mem_used'] = mem_used
        gpu_info[idx]['mem_total'] = mem_total
        gpu_info[idx]['mem_used_percent'] = 100. * mem_used / mem_total

    return gpu_info
#----------------------------------------------------------------------------------------------------------------------------------


def available_gpus(free_memory=4000):
    return [gpu['index'] for gpu in gpu_info() if (gpu['mem_total'] - gpu['mem_used']) >= free_memory]
#----------------------------------------------------------------------------------------------------------------------------------


def wait_available_gpu(free_memory=4000):
    all_gpus = gpu_info()
    if not len(all_gpus):
        log.warning("None GPU available")
        raise Exception("None GPU available")

    # GLOBAL_GPU_LOCK.acquire()
    while True:
        all_gpus = gpu_info()
        gpu_list = sorted([(gpu['index'], gpu['mem_total'] - gpu['mem_used']) for gpu in all_gpus if (gpu['mem_total'] - gpu['mem_used']) >= free_memory], key=lambda x: x[1])
        if len(gpu_list) == 0:
            log.info("GPU wait 30 seconds")
            time.sleep(10)
            continue

        best_gpu = int(gpu_list[-1][0])
        with nvml_context():
            gpu_data = device_status(best_gpu)
            for gpu in all_gpus:
                if int(gpu["index"]) == best_gpu:
                    gpu_data.update(gpu)
                    break

        # GLOBAL_GPU_LOCK.release()
        log.info(f"Best GPU: {gpu_data}")
        return gpu_data
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Dataset manipulation functions

Author: Alex Sobral de Freitas
"""

import json
import random
import copy
import datetime
import dateutil.parser

import unicodedata
import re

import cv2
import numpy as np

from pymongo import MongoClient
from bson.objectid import ObjectId

from eyeflow_sdk.dataset_serialize_pb2 import DsetFile
from eyeflow_sdk.log_obj import log, CONFIG
from eyeflow_sdk.file_access import FileAccess
#----------------------------------------------------------------------------------------------------------------------------------

def slugify(value):
    """
    Normalizes string, removes non-alpha characters, and converts spaces to underscore.
    """
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode()
    value = str(re.sub(r'[^\w\s-]', '', value).strip())
    value = str(re.sub(r'[\s]+', '_', value))
    return value
#----------------------------------------------------------------------------------------------------------------------------------

def default_json_converter(obj):
    if isinstance(obj, datetime.datetime):
        return obj.replace(tzinfo=datetime.timezone.utc).isoformat()
    else:
        return str(obj)
#----------------------------------------------------------------------------------------------------------------------------------


class Dataset():
    """
    Class to serialize and deserialize datasets to/from a file
    """
    def __init__(self, dataset, db_config=None, cloud_parms=None):
        if dataset != slugify(dataset):
            raise Exception("Invalid dataset name %s - Sugestion: %s" % (dataset, slugify(dataset)))

        self.id = None
        self.dataset_name = None

        if len(dataset) == 24:
            try:
                self.id = str(ObjectId(dataset))
            except:
                self.dataset_name = dataset
                self.id = None
        else:
            self.dataset_name = dataset

        if db_config is not None:
            self.db_config = db_config
        else:
            self.db_config = CONFIG["db-service"]

        self._cloud_parms = cloud_parms

        self.parms = {}
        self.examples = []
        self.images = {}


    @staticmethod
    def get_mongo_database(db_config):
        """
        Returns a client to mongo_db access
        """
        client = MongoClient(db_config["db_url"])
        return client[db_config["db_name"]]


    def load_data(self):
        """
        Load dataset data from database
        """
        db_mongo = Dataset.get_mongo_database(self.db_config)

        if self.id:
            self.parms = db_mongo.dataset.find_one({"_id": self.id})

            if self.parms is None:
                raise Exception('Dataset not found: %s' % self.id)

            self.dataset_name = str(self.parms["name"])
        else:
            self.parms = db_mongo.dataset.find_one({"name": self.dataset_name})

            if self.parms is None:
                raise Exception('Dataset not found: %s' % self.dataset_name)

            self.id = str(self.parms["_id"])

        cursor = db_mongo.example.find({"dataset_id": self.parms["_id"]})
        if cursor is not None:
            for exp in cursor:
                self.examples.append(exp)

        self.load_images_from_disk()
        log.info("Load dataset from database %d examples" % len(self.examples))


    def load_images_from_disk(self, origin="cloud"):
        """
        Load images from disk
        """
        file_ac = FileAccess(storage="dataset", resource_id=self.id, cloud_parms=self._cloud_parms)
        if self._cloud_parms is not None:
            file_ac.sync_files(origin=origin)

        for exp in self.examples:
            with file_ac.open(exp["example"], 'rb') as fp:
                img = fp.read()
                self.images[exp["example"]] = img


    def update_default_parms(self):
        """
        Performs an update in parms with all default parms read from database
        """
        db_mongo = Dataset.get_mongo_database(self.db_config)

        network_parms = {}
        dataset_network_parms = self.parms.get("network_parms", network_parms)
        network_default_parms = db_mongo.parms.find_one({"name": "network_default_parms"})
        if network_default_parms is not None:
            network_parms.update(network_default_parms["network_parms"][self.parms["info"]["type"]])

        dataset_default_parms = db_mongo.parms.find_one({"name": "dataset_default_parms"})
        if dataset_default_parms is not None:
            network_parms.update(dataset_default_parms["network_parms"][self.parms["info"]["type"]])

        network_parms.update(dataset_network_parms)

        self.parms.update({"network_parms": network_parms})


    def save_data_db(self):
        """
        Save dataset data to database
        """
        db_mongo = Dataset.get_mongo_database(self.db_config)

        dataset_collection = db_mongo.dataset
        dataset_collection.delete_many({"name": self.parms['name']})

        self.parms["info"]['modified_date'] = datetime.datetime.now()
        self.parms["_id"] = ObjectId(self.id)

        dataset_collection.delete_many({"_id": self.parms['_id']})
        dataset_collection.insert_one(self.parms)

        example_collection = db_mongo.example
        example_collection.delete_many({"dataset": self.parms['name']})
        example_collection.delete_many({"dataset_id": self.parms['_id']})

        for exp in self.examples:
            exp["dataset"] = self.parms['name']
            exp["dataset_id"] = self.parms['_id']
            example_collection.insert_one(exp)

        self.save_images_to_disk()

        log.info("Inserted dataset into database %d examples" % len(self.examples))


    def save_images_to_disk(self, origin="local"):
        """
        Save images to disk
        """
        file_ac = FileAccess(storage="dataset", resource_id=self.id, cloud_parms=self._cloud_parms)

        for exp in self.examples:
            if file_ac.is_file(exp["example"]):
                file_ac.remove_file(exp["example"])

            with file_ac.open(exp["example"], 'wb') as fp:
                fp.write(self.images[exp["example"]])

        if self._cloud_parms is not None:
            file_ac.sync_files(origin=origin)


    def get_example_img(self, example_id):
        """
        Returns an image in opencv format
        """
        if example_id not in self.images:
            raise Exception(f'Example {example_id} not found')

        return cv2.imdecode(np.frombuffer(self.images[example_id], dtype=np.uint8), cv2.IMREAD_COLOR)


    def get_train_subsets(self):
        """
        Partitioning of dataset in 3 groups: train, validation and test
        """
        if not self.parms:
            raise Exception('Must load or import dataset first')

        train_dataset = Dataset(self.id)
        train_dataset.parms = copy.deepcopy(self.parms)
        val_dataset = Dataset(self.id)
        val_dataset.parms = copy.deepcopy(self.parms)
        test_dataset = Dataset(self.id)
        test_dataset.parms = copy.deepcopy(self.parms)

        random.shuffle(self.examples)

        val_size = int(len(self.examples) * self.parms["network_parms"]["val_size"])
        if val_size < 1:
            val_size = 1

        test_size = int(len(self.examples) * self.parms["network_parms"]["test_size"])
        if test_size < 1:
            test_size = 1

        if (val_size + test_size) > len(self.examples):
            raise Exception("Insufficient examples to train: %d" % len(self.examples))

        train_dataset.examples = copy.deepcopy(self.examples[test_size + val_size:])
        val_dataset.examples = copy.deepcopy(self.examples[test_size:test_size + val_size])
        test_dataset.examples = copy.deepcopy(self.examples[:test_size])

        def copy_imgs(dataset, img_list):
            for exp in dataset.examples:
                dataset.images[exp["example"]] = copy.deepcopy(img_list[exp["example"]])

        copy_imgs(train_dataset, self.images)
        copy_imgs(val_dataset, self.images)
        copy_imgs(test_dataset, self.images)

        return train_dataset, val_dataset, test_dataset


    @staticmethod
    def export_to_pb(parms, examples, images, filename):
        """
        Export dataset to file in protobuf format
        """
        dsetfile = DsetFile()
        dsetfile.parms = json.dumps(parms, ensure_ascii=False, default=default_json_converter)
        dsetfile.examples.extend([json.dumps(exp, ensure_ascii=False, default=default_json_converter) for exp in examples])
        for img in images:
            dsetfile.images[img] = images[img]

        file_ac = FileAccess(storage="export")

        with file_ac.open(filename, 'wb') as fp:
            fp.write(dsetfile.SerializeToString())


    def export_dataset(self):
        """
        Export dataset to folder adding only differences if a base export exists
        """

        file_ac = FileAccess(storage="export", cloud_parms=self._cloud_parms)
        base_file_name = self.id + ".dset"
        diff_file_name = self.id + '_diff.dset'
        data_filename = self.id + ".json"

        def save_export_data(filename, diff_examples):
            export_data = {
                "export_version": "1",
                "export_date": datetime.datetime.now(),
                "dataset": self.dataset_name,
                "dataset_id": self.id,
                "num_examples": len(self.examples),
                "diff_examples": diff_examples,
                "images_list": list(self.images.keys())
            }

            if file_ac.is_file(filename):
                file_ac.remove_file(filename)

            with file_ac.open(filename, 'w', newline='', encoding='utf8') as fp:
                json.dump(export_data, fp, ensure_ascii=False, indent=2, default=default_json_converter)


        if not file_ac.is_file(base_file_name):
            # Dataset.export_to_file(base_file_name)
            Dataset.export_to_pb(self.parms, self.examples, self.images, base_file_name)
            save_export_data(data_filename, diff_examples=0)
            if self._cloud_parms is not None:
                file_ac.sync_files(file_list=[base_file_name, data_filename],
                                   origin="local")
            log.info("Dataset exported to file %d examples" % len(self.examples))
            return

        base_parms, base_examples = Dataset.import_from_pb(base_file_name, retrieve_images=False)

        base_examples_list = {}
        for exp in base_examples:
            base_examples_list[str(exp["_id"])] = exp

        export_examples = []
        export_images = {}
        for exp in self.examples:
            if str(exp["_id"]) not in base_examples_list:
                export_examples.append(exp)
                export_images[exp["example"]] = self.images[exp["example"]]
            elif exp != base_examples_list[str(exp["_id"])]:
                export_examples.append(exp)

        if export_examples or (self.parms != base_parms):
            Dataset.export_to_pb(self.parms, export_examples, export_images, diff_file_name)
            save_export_data(data_filename, diff_examples=len(export_examples))
            if self._cloud_parms is not None:
                file_ac.sync_files(file_list=[base_file_name, diff_file_name, data_filename],
                                   origin="local")

        log.info("Dataset exported to file %d examples" % len(self.examples))


    @staticmethod
    def import_from_pb(filename, retrieve_images=True):
        """
        Import dataset from file in protobuf format
        """
        def convert_date(str_date):
            return dateutil.parser.isoparse(str_date)

        file_ac = FileAccess(storage="export")

        dset = DsetFile()

        with file_ac.open(filename, 'rb') as fp:
            dset.ParseFromString(fp.read())

        parms = json.loads(dset.parms)
        parms["_id"] = ObjectId(parms["_id"])
        parms["info"]["creation_date"] = convert_date(parms["info"].get("creation_date"))
        parms["info"]["modified_date"] = convert_date(parms["info"].get("modified_date"))

        examples = []
        for exp in dset.examples:
            exp = json.loads(exp)
            exp["_id"] = ObjectId(exp["_id"])
            exp["date"] = convert_date(exp.get("date"))
            if "modified_date" in exp:
                exp["modified_date"] = convert_date(exp.get("modified_date"))

            examples.append(exp)

        if not retrieve_images:
            return parms, examples

        images = {}
        for img in dset.images:
            images[img] = dset.images[img]

        return parms, examples, images


    def import_dataset(self):
        """
        Import dataset from folder with base and diffs
        """

        file_list = []
        if self.id:
            file_list.append(self.id + ".dset")
            file_list.append(self.id + '_diff.dset')
            file_list.append(self.id + ".json")

        if self.dataset_name:
            file_list.append(self.dataset_name + ".dset")
            file_list.append(self.dataset_name + '_diff.dset')
            file_list.append(self.dataset_name + ".json")

        file_ac = FileAccess(storage="export", cloud_parms=self._cloud_parms)

        if self._cloud_parms is not None:
            file_ac.sync_files(file_list=file_list,
                               origin="cloud"
                              )

        if self.dataset_name and file_ac.is_file(f"{self.dataset_name}.json"):
            base_file_name = self.dataset_name + ".dset"
            diff_file_name = self.dataset_name + '_diff.dset'
            data_filename = self.dataset_name + ".json"
        elif self.id and file_ac.is_file(f"{self.id}.json"):
            base_file_name = self.id + ".dset"
            diff_file_name = self.id + '_diff.dset'
            data_filename = self.id + ".json"
        else:
            raise Exception(f'Base export file {self.dataset_name}.json or {self.id}.json does not exist')

        with file_ac.open(data_filename, 'r', newline='', encoding='utf8') as fp:
            export_data = json.load(fp)

        if not file_ac.is_file(base_file_name):
            raise Exception(f"Base export file {base_file_name} does not exist")

        self.parms, self.examples, self.images = Dataset.import_from_pb(base_file_name)
        self.dataset_name = str(self.parms["name"])
        self.id = str(self.parms["_id"])

        if file_ac.is_file(diff_file_name) and export_data.get("diff_examples", 0) > 0:
            self.parms, diff_examples, diff_images = Dataset.import_from_pb(diff_file_name)

            for diff_exp in diff_examples:
                for idx, exp in enumerate(self.examples):
                    if str(exp["_id"]) == str(diff_exp["_id"]):
                        self.examples[idx] = diff_exp
                        break
                else:
                    self.examples.append(diff_exp)

            self.images.update(diff_images)

        if len(self.examples) != export_data["num_examples"]:
            raise Exception(f"Invalid number of examples. Imported: {len(self.examples)} - Exported: {export_data['num_examples']}")

        if len(self.images) != len(export_data["images_list"]):
            raise Exception(f"Invalid number of images. Imported: {len(self.images)} - Exported: {len(export_data['images_list'])}")

        log.info("Dataset imported from file %d examples" % len(self.examples))
# ---------------------------------------------------------------------------------------------------------------------------------
